#include <sys/types.h>
#include <stdbool.h>
#include <sys/event.h>
#include <c.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdio.h>
#include <sys/errno.h>

//We will not allocate nor read more than this many bytes, even if we could. This keeps us from trying to allocate, say, 2 GiB of memory.
#ifndef MAX_READ_CHUNK
#	define MAX_READ_CHUNK 104857600U
#endif

int main(int argc, const char **argv) {
	fprintf(stderr, "This program will exit when event.flags & 0x%x or when read returns <= 0.\n\n", EV_EOF);

	int read_fd = STDIN_FILENO;
	if(argc > 1) {
		fprintf(stderr, "Will read directly from %s.\n", argv[1]);
		read_fd = open(argv[1], O_RDONLY, 0644);
	} else {
		fputs("Will read from stdin.\n", stderr);
		/*
		//Set the stdin FD as non-blocking.
		int flags = fcntl(read_fd, F_GETFL, 0);
		flags |= O_NONBLOCK;
		fcntl(read_fd, F_SETFL, flags);
		 */
	}

	//Fill out an event structure and add it.
	int q = kqueue();

	struct kevent event_to_add = {
		.ident = read_fd,
		.filter = EVFILT_READ,
		.flags = EV_ADD,
		.fflags = 0U,
		.data = 0U,
		.udata = NULL
	};
	kevent(q, &event_to_add, 1, NULL, 0, /*timeout*/ NULL);

	//Make absolutely sure that this isn't a problem of reused storage.
	bzero(&event_to_add, sizeof(event_to_add));
	fprintf(stderr, "event_to_add.flags: 0x%x.\n", event_to_add.flags);

	char *buf = NULL;
	size_t buflen = 0;

	//This same program also tests another bug (event.data having irrational values when the amount of data available to be read exceeds LONG_MAX bytes). A second variable exists here in that version, to make the loop run one more time after read returns <= 0.
	bool still_getting_data = true;

	while(still_getting_data) {
		struct kevent event;
		int got_event = kevent(q, NULL, 0, &event, 1, NULL);

		if(event.filter == EVFILT_READ) {
			fprintf(stderr, "Got read event with event.data (number of bytes available to read) = %li (%lu); event.flags = 0x%x.\n", event.data, event.data, event.flags);
			if(event.data) {
				//The first time, realloc is passed NULL, and so acts as malloc.
				buflen = event.data;

				//To be safe, we will not read more than 1 MiB at a time.
				if(buflen > MAX_READ_CHUNK) {
					buflen = MAX_READ_CHUNK;
					fprintf(stderr, "Note: Read event offered %li (%lu) bytes. Throttled down to %u.\n", event.data, event.data, MAX_READ_CHUNK);
				}

				buf = realloc(buf, buflen);
				if(buf) {
					ssize_t amt_read = read(event.ident, buf, buflen);
					fprintf(stderr, "Read %i bytes from FD %i.\n", amt_read, event.ident);
					if(amt_read <= 0) {
						if(amt_read < 0)
							fprintf(stderr, "Got error: %s\n", strerror(errno));

						still_getting_data = false;
					}
				} else
					fprintf(stderr, "WARNING: could not allocate %zu bytes. Nothing was read.\n", buflen);
			}
			if(event.flags & EV_EOF)
				still_getting_data = false;

			if(!still_getting_data) {
				fputs("End of loop.\n", stderr);

				//Closing our stdin means that we'll no longer get read events.
				close(read_fd);
			}
		}

		close(q);
	}

	free(buf);
	close(read_fd);

	return 0;
}
